from datetime import date, datetime, time
from typing import Any, Dict, List


def to_toml(data: Dict[str, Any]) -> str:
    """Convert a dictionary to TOML format string."""
    lines = []

    # Handle top-level key-value pairs first
    top_level = {}
    tables = {}

    for key, value in data.items():
        if isinstance(value, dict):
            tables[key] = value
        else:
            top_level[key] = value

    # Write top-level pairs
    for key, value in top_level.items():
        lines.append(f"{key} = {_format_value(value)}")

    # Write tables
    for table_name, table_data in tables.items():
        if lines:  # Add blank line before tables
            lines.append("")
        lines.append(f"[{table_name}]")
        _write_table(lines, table_data)

    return "\n".join(lines)


def _format_value(value: Any) -> str:
    """Format a Python value as TOML."""
    if isinstance(value, str):
        # Escape special characters and quote
        escaped = value.replace("\\", "\\\\").replace('"', '\\"')
        return f'"{escaped}"'
    elif isinstance(value, bool):
        return "true" if value else "false"
    elif isinstance(value, (int, float)):
        return str(value)
    elif isinstance(value, datetime):
        return value.isoformat()
    elif isinstance(value, date):
        return value.isoformat()
    elif isinstance(value, time):
        return value.isoformat()
    elif isinstance(value, list):
        formatted_items = [_format_value(item) for item in value]
        return f"[{', '.join(formatted_items)}]"
    elif value is None:
        raise ValueError("TOML doesn't support null values")
    else:
        raise TypeError(f"Unsupported type: {type(value)}")


def _write_table(lines: List[str], data: Dict[str, Any]) -> None:
    """Write table contents."""
    for key, value in data.items():
        if not isinstance(value, dict):
            lines.append(f"{key} = {_format_value(value)}")



from logging import getLogger

from inspect_ai.util import trace_message

logger = getLogger(__file__)


def trace(message: str) -> None:
    logger.setLevel("TRACE")
    trace_message(logger, category="Inspect SWE", message=message)



import json
from pathlib import Path
from typing import Any

from typing_extensions import Literal

from .._util.agentbinary import AgentBinarySource, AgentBinaryVersion
from .._util.appdirs import package_cache_dir
from .._util.download import download_text_file
from .._util.sandbox import SandboxPlatform
from .._util.tarball import extract_tarball


def codex_cli_binary_source() -> AgentBinarySource:
    cached_binary_dir = package_cache_dir("codex-cli-downloads")

    async def resolve_version(
        version: Literal["stable", "latest"] | str, platform: SandboxPlatform
    ) -> AgentBinaryVersion:
        # Resolve version alias if needed
        if version in ["stable", "latest"]:
            version = await _fetch_latest_stable_version()

        # Get release information
        release = await _fetch_release_assets(version)

        # Get the platform-specific asset
        arch = _platform_to_codex_arch(platform)
        asset_name = f"codex-{arch}.tar.gz"

        # Find the matching asset
        asset = None
        for a in release.get("assets", []):
            if a["name"] == asset_name:
                asset = a
                break

        if asset is None:
            raise RuntimeError(
                f"No asset found for platform {platform} in version {version}"
            )

        # Extract checksum (format: "sha256:xxx")
        digest = asset.get("digest", "")
        if not digest.startswith("sha256:"):
            raise RuntimeError(f"Invalid digest format: {digest}")
        expected_checksum = digest[7:]  # Remove "sha256:" prefix

        # Get download URL
        download_url = asset["browser_download_url"]

        return AgentBinaryVersion(version, expected_checksum, download_url)

    def cached_binary_path(version: str, platform: SandboxPlatform) -> Path:
        return cached_binary_dir / f"codex-{version}-{platform}"

    def list_cached_binaries() -> list[Path]:
        return list(cached_binary_dir.glob("codex-*"))

    return AgentBinarySource(
        agent="codex cli",
        binary="codex",
        resolve_version=resolve_version,
        cached_binary_path=cached_binary_path,
        list_cached_binaries=list_cached_binaries,
        post_download=extract_tarball,
        post_install=None,
    )


def _platform_to_codex_arch(platform: SandboxPlatform) -> str:
    """Map SandboxPlatform to Codex architecture string.

    Always use musl variants for better compatibility since they're
    statically linked and don't depend on system GLIBC version.
    """
    platform_map = {
        "linux-x64": "x86_64-unknown-linux-musl",
        "linux-x64-musl": "x86_64-unknown-linux-musl",
        "linux-arm64": "aarch64-unknown-linux-musl",
        "linux-arm64-musl": "aarch64-unknown-linux-musl",
    }
    if platform not in platform_map:
        raise ValueError(f"Unsupported platform: {platform}")
    return platform_map[platform]


async def _fetch_latest_stable_version() -> str:
    """Fetch the latest stable version from GitHub releases."""
    releases_url = "https://api.github.com/repos/openai/codex/releases"
    releases_json = await download_text_file(releases_url)
    releases = json.loads(releases_json)

    # Filter out pre-releases and alpha versions
    stable_releases = [
        r
        for r in releases
        if not r.get("prerelease", False) and "-alpha" not in r.get("tag_name", "")
    ]

    if not stable_releases:
        raise RuntimeError("No stable releases found for codex")

    # Get the most recent stable release
    latest = stable_releases[0]
    tag_name = latest["tag_name"]

    # Extract version from tag (e.g., "rust-v0.29.0" -> "0.29.0")
    if tag_name.startswith("rust-v"):
        result: str = tag_name[6:]  # Remove "rust-v" prefix
        return result
    else:
        raise RuntimeError(f"Unexpected tag format: {tag_name}")


async def _fetch_release_assets(version: str) -> dict[str, Any]:
    """Fetch release assets for a specific version."""
    tag = f"rust-v{version}"
    release_url = f"https://api.github.com/repos/openai/codex/releases/tags/{tag}"
    release_json = await download_text_file(release_url)
    result: dict[str, Any] = json.loads(release_json)
    return result



from dataclasses import dataclass
from pathlib import Path
from typing import Awaitable, Callable, Literal, NamedTuple

from inspect_ai.util import SandboxEnvironment, concurrency
from inspect_ai.util import sandbox as sandbox_env

from inspect_swe._util.trace import trace

from .checksum import verify_checksum
from .download import download_file
from .sandbox import (
    SandboxPlatform,
    bash_command,
    detect_sandbox_platform,
    sandbox_exec,
)


class AgentBinaryVersion(NamedTuple):
    version: str
    expected_checksum: str
    download_url: str


@dataclass
class AgentBinarySource:
    agent: str
    binary: str
    resolve_version: Callable[
        [Literal["stable", "latest"] | str, SandboxPlatform],
        Awaitable[AgentBinaryVersion],
    ]
    cached_binary_path: Callable[[str, SandboxPlatform], Path]
    list_cached_binaries: Callable[[], list[Path]]
    post_download: Callable[[bytes], bytes] | None
    post_install: str | None


async def ensure_agent_binary_installed(
    source: AgentBinarySource,
    version: Literal["auto", "sandbox", "stable", "latest"] | str = "auto",
    user: str | None = None,
    sandbox: SandboxEnvironment | None = None,
) -> str:
    # resolve sandbox
    sandbox = sandbox or sandbox_env()

    # look in the sandbox first if we need to
    if version == "auto" or version == "sandbox":
        result = await sandbox.exec(bash_command(f"which {source.binary}"), user=user)
        if result.success:
            binary_path = result.stdout.strip()
            trace(f"Using {source.agent} installed in sandbox: {binary_path}")
            return binary_path

        # if version == "sandbox" and we don't find it that's an error
        if version == "sandbox":
            raise RuntimeError(f"unable to locate {source.agent} in sandbox")

        # otherwise set to "stable"
        version = "stable"

    # detect the sandbox target platform
    platform = await detect_sandbox_platform(sandbox)

    # use concurrency so multiple samples don't attempt the same download all at once
    async with concurrency(f"{source.binary}-install", 1, visible=False):
        # if a specific version is requested, first try to read it directly from the cache
        if version not in ["stable", "latest"]:
            binary_bytes: bytes | None = read_cached_binary(
                source, version, platform, None
            )
            if binary_bytes is not None:
                trace(f"Used claude code binary from cache: {version} ({platform})")
        else:
            binary_bytes = None

        # download the binary
        if binary_bytes is None:
            binary_bytes, resolved_version = await download_agent_binary_async(
                source, version, platform, trace
            )
        else:
            # If we got it from cache, version is already the resolved version
            resolved_version = version

        # write it into the container and return it
        binary_path = (
            f"/var/tmp/.5c95f967ca830048/{source.binary}-{resolved_version}-{platform}"
        )
        await sandbox.write_file(binary_path, binary_bytes)
        await sandbox_exec(sandbox, f"chmod +x {binary_path}", user="root")
        if source.post_install:
            await sandbox_exec(
                sandbox, f"{binary_path} {source.post_install}", user=user
            )
        return binary_path


async def download_agent_binary_async(
    source: AgentBinarySource,
    version: Literal["stable", "latest"] | str,
    platform: SandboxPlatform,
    logger: Callable[[str], None] | None = None,
) -> tuple[bytes, str]:
    # resovle logger
    logger = logger or print

    # determine version and checksum
    version, expected_checksum, download_url = await source.resolve_version(
        version, platform
    )

    # check the cache (if post_download is used, don't verify checksum since cached is processed)
    cache_checksum = None if source.post_download else expected_checksum
    binary_data = read_cached_binary(source, version, platform, cache_checksum)
    if binary_data is None:
        # not in cache, download and verify checksum
        binary_data = await download_file(download_url)
        if not verify_checksum(binary_data, expected_checksum):
            raise ValueError("Checksum verification failed")

        # apply post-download processing if provided (e.g., extract from tar.gz)
        if source.post_download is not None:
            binary_data = source.post_download(binary_data)

        # save to cache
        write_cached_binary(source, binary_data, version, platform)

        # trace
        logger(f"Downloaded {source.agent} binary: {version} ({platform})")
    else:
        logger(f"Used {source.agent} binary from cache: {version} ({platform})")

    # return data and resolved version
    return binary_data, version


def read_cached_binary(
    source: AgentBinarySource,
    version: str,
    platform: SandboxPlatform,
    expected_checksum: str | None,
) -> bytes | None:
    # no cached binary
    cache_path = source.cached_binary_path(version, platform)
    if not cache_path.exists():
        return None

    # read binary
    with open(cache_path, "rb") as f:
        binary_data = f.read()

    if expected_checksum is None or verify_checksum(binary_data, expected_checksum):
        cache_path.touch()
        return binary_data
    else:
        cache_path.unlink()
        return None


def write_cached_binary(
    source: AgentBinarySource,
    binary_data: bytes,
    version: str,
    platform: SandboxPlatform,
) -> None:
    binary_path = source.cached_binary_path(version, platform)

    with open(binary_path, "wb") as f:
        f.write(binary_data)

    _cleanup_binary_cache(source, keep_count=3)


def _cleanup_binary_cache(source: AgentBinarySource, keep_count: int = 5) -> None:
    # get all cached binaries
    cache_files = source.list_cached_binaries()
    if len(cache_files) <= keep_count:
        return

    # remove oldest
    cache_files.sort(key=lambda f: f.stat().st_atime)
    files_to_remove = cache_files[:-keep_count]
    for file_path in files_to_remove:
        file_path.unlink()



from typing import no_type_check


@no_type_check
def running_in_notebook() -> bool:
    try:
        from IPython import get_ipython  # type: ignore

        if "IPKernelApp" not in get_ipython().config:
            return False
    except ImportError:
        return False
    except AttributeError:
        return False
    return True



import re

import yaml


def read_front_matter_name(content: str) -> str | None:
    # front-matter
    frontmatter_match = re.match(r"^\s*---\s*\n(.*?)\n---", content, re.DOTALL)
    if not frontmatter_match:
        return None
    frontmatter = frontmatter_match.group(1)

    try:
        # Parse as YAML
        data = yaml.safe_load(frontmatter)
        if "name" in data:
            return str(data.get("name"))
        else:
            return None
    except yaml.YAMLError:
        return None



import httpx


async def download_file(url: str) -> bytes:
    async with httpx.AsyncClient() as client:
        response = await client.get(url, follow_redirects=True)
        response.raise_for_status()
        return response.content


async def download_text_file(url: str) -> str:
    return (await download_file(url)).decode("utf-8")



import shlex
from logging import getLogger
from pathlib import Path
from textwrap import dedent
from typing import Any, Literal, Sequence

from inspect_ai.agent import (
    Agent,
    AgentAttempts,
    AgentState,
    BridgedToolsSpec,
    agent,
    agent_with,
    sandbox_agent_bridge,
)
from inspect_ai.model import ChatMessageSystem, GenerateFilter
from inspect_ai.scorer import score
from inspect_ai.tool import MCPServerConfig, Skill, install_skills, read_skills
from inspect_ai.util import SandboxEnvironment, store
from inspect_ai.util import sandbox as sandbox_env

from inspect_swe._util._async import is_callable_coroutine
from inspect_swe._util.centaur import CentaurOptions, run_centaur
from inspect_swe._util.messages import build_user_prompt
from inspect_swe._util.path import join_path
from inspect_swe._util.sandbox import sandbox_exec
from inspect_swe._util.toml import to_toml
from inspect_swe._util.trace import trace

from .._util.agentbinary import ensure_agent_binary_installed
from .agentbinary import codex_cli_binary_source

logger = getLogger(__file__)


@agent
def codex_cli(
    name: str = "Codex CLI",
    description: str = dedent("""
       Autonomous coding agent capable of writing, testing, debugging,
       and iterating on code across multiple languages.
    """),
    system_prompt: str | None = None,
    model_config: str = "gpt-5.1",
    skills: Sequence[str | Path | Skill] | None = None,
    mcp_servers: Sequence[MCPServerConfig] | None = None,
    bridged_tools: Sequence[BridgedToolsSpec] | None = None,
    disallowed_tools: list[Literal["web_search"]] | None = None,
    centaur: bool | CentaurOptions = False,
    attempts: int | AgentAttempts = 1,
    model: str | None = None,
    filter: GenerateFilter | None = None,
    retry_refusals: int | None = None,
    home_dir: str | None = None,
    cwd: str | None = None,
    env: dict[str, str] | None = None,
    user: str | None = None,
    sandbox: str | None = None,
    version: Literal["auto", "sandbox", "latest"] | str = "auto",
    config_overrides: dict[str, str] | None = None,
) -> Agent:
    """Codex CLI.

    Agent that uses OpenAI [Codex CLI](https://github.com/openai/codex) running in a sandbox.

    Use the `attempts` option to enable additional submissions if the initial
    submission(s) are incorrect (by default, no additional attempts are permitted).

    Args:
        name: Agent name (used in multi-agent systems with `as_tool()` and `handoff()`)
        description: Agent description (used in multi-agent systems with `as_tool()` and `handoff()`)
        system_prompt: Additional system prompt to append to default system prompt.
        model_config: Model configuration profile (e.g. used to determine the system prompt).
        skills: Additional [skills](https://inspect.aisi.org.uk/tools-standard.html#sec-skill) to make available to the agent.
        mcp_servers: MCP servers to make available to the agent.
        bridged_tools: Host-side Inspect tools to expose to the agent via MCP.
            Each BridgedToolsSpec creates an MCP server that makes the specified
            tools available to the agent running in the sandbox.
        disallowed_tools: Optionally disallow tools (currently only web_search).
        centaur: Run in 'centaur' mode, which makes Codex CLI available to an Inspect `human_cli()` agent rather than running it unattended.
        attempts: Configure agent to make multiple attempts.
        model: Model name to use (defaults to main model for task).
        filter: Filter for intercepting bridged model requests.
        retry_refusals: Should refusals be retried? (pass number of times to retry)
        home_dir: Home directory to use for codex cli. If set, AGENTS.md, skills, and the MCP configuration will be written here.
        cwd: Working directory to run codex cli within.
        env: Environment variables to set for codex cli
        user: User to execute codex cli with.
        sandbox: Optional sandbox environment name.
        version: Version of codex cli to use. One of:
            - "auto": Use any available version of codex cli in the sandbox, otherwise download the latest version.
            - "sandbox": Use the version of codex cli in the sandbox (raises `RuntimeError` if codex is not available in the sandbox)
            - "latest": Download and use the very latest version of codex cli.
            - "x.x.x": Download and use a specific version of codex cli.
        config_overrides: Additional Codex CLI configuration overrides.
            Each key-value pair is passed as `-c key=value` to the CLI.
    """
    # resolve centaur
    if centaur is True:
        centaur = CentaurOptions()

    # resolve model
    model = f"inspect/{model}" if model is not None else "inspect"

    # resolve skills
    resolved_skills = read_skills(skills) if skills is not None else None

    # resolve attempts
    attempts = AgentAttempts(attempts) if isinstance(attempts, int) else attempts

    # ensure disallowed_tools list
    disallowed_tools = disallowed_tools or []

    async def execute(state: AgentState) -> AgentState:
        # determine port (use new port for each execution of agent on sample)
        MODEL_PORT = "codex_cli_model_port"
        port = store().get(MODEL_PORT, 3000) + 1
        store().set(MODEL_PORT, port)

        async with sandbox_agent_bridge(
            state,
            model=model,
            filter=filter,
            retry_refusals=retry_refusals,
            port=port,
            bridged_tools=bridged_tools,
        ) as bridge:
            # ensure codex is installed and get binary location
            codex_binary = await ensure_agent_binary_installed(
                codex_cli_binary_source(), version, user, sandbox_env(sandbox)
            )

            # build system prompt
            system_messages = [
                m.text for m in state.messages if isinstance(m, ChatMessageSystem)
            ]
            if system_prompt is not None:
                system_messages.append(system_prompt)

            # resolve sandbox
            sbox = sandbox_env(sandbox)

            # determine CODEX_HOME (default to whatever sandbox working dir is)
            if home_dir is None:
                working_dir = await sandbox_exec(sbox, "pwd", user=user, cwd=cwd)
                codex_home = join_path(working_dir, ".codex")
            else:
                # Resolve ~ and $VARS inside the sandbox
                codex_home = await sandbox_exec(
                    sbox, f'eval echo "{home_dir}"', user=user, cwd=cwd
                )
            await sandbox_exec(sbox, cmd=f"mkdir -p {codex_home}", user=user)

            # location for agents_md
            def codex_agents_md() -> str:
                AGENTS_MD = "AGENTS.md"
                if home_dir is not None:
                    return join_path(codex_home, AGENTS_MD)
                elif cwd is not None:
                    return join_path(cwd, AGENTS_MD)
                else:
                    return AGENTS_MD

            # location for config_toml (either codex_home or cwd/.codex )
            async def codex_config_toml() -> str:
                CONFIG_TOML = "config.toml"
                if home_dir is not None:
                    return join_path(codex_home, CONFIG_TOML)
                else:
                    dir = ".codex" if cwd is None else join_path(cwd, ".codex")
                    await sandbox_exec(sbox, cmd=f"mkdir -p {dir}", user=user)
                    return join_path(dir, CONFIG_TOML)

            # write system messages to AGENTS.md
            if system_messages:
                await sbox.write_file(codex_agents_md(), "\n\n".join(system_messages))

            # install skills
            if resolved_skills is not None:
                await install_skills(
                    resolved_skills, sbox, user, join_path(codex_home, "skills")
                )

            prompt, has_assistant_response = build_user_prompt(state.messages)

            # build agent cmd
            cmd = [codex_binary]

            # headless
            if centaur is False:
                cmd.extend(["exec", "--color", "never", "--skip-git-repo-check"])

            # default cli args
            cmd.extend(
                [
                    # real model is passed to the bridge above, this just affects defaults e.g. system prompt
                    "--model",
                    model_config,
                    "--dangerously-bypass-approvals-and-sandbox",
                ]
            )

            # include web search if appropriate
            if "web_search" not in disallowed_tools:
                cmd.extend(["--enable", "web_search_request"])

            # apply config overrides
            if config_overrides:
                for key, value in config_overrides.items():
                    cmd.extend(["-c", f"{key}={value}"])

            # build toml config
            toml_config: dict[str, Any] = {}

            # register mcp servers (combine static configs with bridged tools)
            all_mcp_servers = list(mcp_servers or []) + bridge.mcp_server_configs
            if all_mcp_servers:
                for mcp_server in all_mcp_servers:
                    toml_config[f"mcp_servers.{mcp_server.name}"] = (
                        mcp_server.model_dump(
                            exclude={"name", "tools"}, exclude_none=True
                        )
                    )

            # model provider if we are in centaur mode
            if centaur:
                toml_config["preferred_auth_method"] = "apikey"
                toml_config["model_provider"] = "openai-proxy"
                toml_config["model_providers.openai-proxy"] = {
                    "name": "OpenAI Proxy",
                    "base_url": f"http://localhost:{bridge.port}/v1",
                    "env_key": "OPENAI_API_KEY",
                    "wire_api": "responses",
                }

            # write toml config if we have it
            if len(toml_config) > 0:
                await sbox.write_file(await codex_config_toml(), to_toml(toml_config))

            # setup agent env
            agent_env = {
                "CODEX_HOME": codex_home,
                "OPENAI_API_KEY": "api-key",
                "OPENAI_BASE_URL": f"http://localhost:{bridge.port}/v1",
                "RUST_LOG": "warning",
            } | (env or {})

            if centaur:
                await _run_codex_cli_centaur(
                    options=centaur,
                    codex_cmd=cmd,
                    agent_env=agent_env,
                    state=state,
                )
            else:
                # execute the agent (track debug output)
                debug_output: list[str] = []
                agent_prompt = prompt
                attempt_count = 0
                while True:
                    # append prompt
                    agent_cmd = cmd.copy()
                    agent_cmd.append(agent_prompt)

                    # resume previous conversation
                    if has_assistant_response or attempt_count > 0:
                        agent_cmd.extend(["resume", "--last"])

                    # run agent
                    result = await sbox.exec(
                        cmd=["bash", "-c", 'exec 0<&- "$@"', "bash"] + agent_cmd,
                        cwd=cwd,
                        env=agent_env,
                        user=user,
                    )

                    # record output for debug
                    debug_output.append(result.stdout)
                    debug_output.append(result.stderr)

                    # raise for error
                    if not result.success:
                        raise RuntimeError(
                            f"Error executing codex cli agent: {result.stdout}\n{result.stderr}"
                        )

                    # exit if we are at max_attempts
                    attempt_count += 1
                    if attempt_count >= attempts.attempts:
                        break

                    # score this attempt
                    answer_scores = await score(state)

                    # break if we score 'correct'
                    if attempts.score_value(answer_scores[0].value) == 1.0:
                        break

                    # otherwise update prompt with incorrect message and continue
                    else:
                        if callable(attempts.incorrect_message):
                            if not is_callable_coroutine(attempts.incorrect_message):
                                raise ValueError(
                                    "The incorrect_message function must be async."
                                )
                            agent_prompt = await attempts.incorrect_message(
                                state, answer_scores
                            )
                        else:
                            agent_prompt = attempts.incorrect_message

                # trace debug info
                debug_output.insert(0, "Codex CLI Debug Output:")
                trace("\n".join(debug_output))

        # return success
        return bridge.state

    return agent_with(execute, name=name, description=description)


async def _run_codex_cli_centaur(
    options: CentaurOptions,
    codex_cmd: list[str],
    agent_env: dict[str, str],
    state: AgentState,
) -> None:
    instructions = "Codex CLI:\n\n - You may also use Codex CLI via the 'codex' command.\n - Use 'codex resume' if you need to resume a previous codex session."

    # build .bashrc content
    agent_env_vars = [f'export {k}="{v}"' for k, v in agent_env.items()]
    alias_cmd = shlex.join(codex_cmd)
    alias_cmd = "alias codex='" + alias_cmd.replace("'", "'\\''") + "'"
    bashrc = "\n".join(agent_env_vars + ["", alias_cmd])

    # run the human cli
    await run_centaur(options, instructions, bashrc, state)


async def _last_rollout(
    sandbox: SandboxEnvironment, codex_home: str, user: str | None
) -> str | None:
    try:
        rollout = await sandbox_exec(
            sandbox,
            f"find '{codex_home}/sessions' -type f -name 'rollout-*.jsonl' -exec ls -t -- {{}} + | head -n 1",
            user=user,
        )
        return rollout
    except RuntimeError as ex:
        logger.warning(f"Error attempting to read rollout file: {ex}")
        return None



from typing import Literal, TypeAlias, cast

from inspect_ai.util import SandboxEnvironment

SandboxPlatform: TypeAlias = Literal[
    "linux-x64", "linux-arm64", "linux-x64-musl", "linux-arm64-musl"
]


async def detect_sandbox_platform(sandbox: SandboxEnvironment) -> SandboxPlatform:
    # Get OS
    os_name = await sandbox_exec(sandbox, "uname -s")
    if os_name == "Linux":
        os_type = "linux"
    else:
        raise ValueError(f"Unsupported OS: {os_name}")

    # Get architecture
    arch = await sandbox_exec(sandbox, "uname -m")
    if arch in ["x86_64", "amd64"]:
        arch_type = "x64"
    elif arch in ["arm64", "aarch64"]:
        arch_type = "arm64"
    else:
        raise ValueError(f"Unsupported architecture: {arch}")

    # Check for musl on Linux
    if os_type == "linux":
        # Check for musl libc
        musl_check_cmd = (
            "if [ -f /lib/libc.musl-x86_64.so.1 ] || "
            "[ -f /lib/libc.musl-aarch64.so.1 ] || "
            "ldd /bin/ls 2>&1 | grep -q musl; then "
            "echo 'musl'; else echo 'glibc'; fi"
        )
        libc_type = await sandbox_exec(sandbox, musl_check_cmd)
        if libc_type == "musl":
            platform = f"linux-{arch_type}-musl"
        else:
            platform = f"linux-{arch_type}"
    else:
        platform = f"{os_type}-{arch_type}"

    return cast(SandboxPlatform, platform)


def bash_command(cmd: str) -> list[str]:
    return ["bash", "-c", cmd]


async def sandbox_exec(
    sandbox: SandboxEnvironment,
    cmd: str,
    user: str | None = None,
    cwd: str | None = None,
) -> str:
    result = await sandbox.exec(bash_command(cmd), user=user, cwd=cwd)
    if not result.success:
        raise RuntimeError(f"Error executing sandbox command {cmd}: {result.stderr}")
    return result.stdout.strip()



import shlex
import uuid
from pathlib import Path
from textwrap import dedent
from typing import Any, Literal, Sequence

import anyio
from inspect_ai.agent import (
    Agent,
    AgentAttempts,
    AgentState,
    BridgedToolsSpec,
    agent,
    agent_with,
    sandbox_agent_bridge,
)
from inspect_ai.model import ChatMessageSystem, GenerateFilter
from inspect_ai.scorer import score
from inspect_ai.tool import MCPServerConfig, Skill, install_skills, read_skills
from inspect_ai.util import sandbox as sandbox_env
from inspect_ai.util import store
from pydantic_core import to_json

from inspect_swe._util.centaur import CentaurOptions, run_centaur
from inspect_swe._util.path import join_path

from .._util._async import is_callable_coroutine
from .._util.agentbinary import ensure_agent_binary_installed
from .._util.messages import build_user_prompt
from .._util.trace import trace
from .agentbinary import claude_code_binary_source


@agent
def claude_code(
    name: str = "Claude Code",
    description: str = dedent("""
       Autonomous coding agent capable of writing, testing, debugging,
       and iterating on code across multiple languages.
    """),
    system_prompt: str | None = None,
    skills: Sequence[str | Path | Skill] | None = None,
    mcp_servers: Sequence[MCPServerConfig] | None = None,
    bridged_tools: Sequence[BridgedToolsSpec] | None = None,
    disallowed_tools: list[str] | None = None,
    centaur: bool | CentaurOptions = False,
    attempts: int | AgentAttempts = 1,
    model: str | None = None,
    opus_model: str | None = None,
    sonnet_model: str | None = None,
    haiku_model: str | None = None,
    subagent_model: str | None = None,
    filter: GenerateFilter | None = None,
    retry_refusals: int | None = None,
    retry_timeouts: int | None = None,
    cwd: str | None = None,
    env: dict[str, str] | None = None,
    user: str | None = None,
    sandbox: str | None = None,
    version: Literal["auto", "sandbox", "stable", "latest"] | str = "auto",
) -> Agent:
    """Claude Code agent.

    Agent that uses [Claude Code](https://docs.anthropic.com/en/docs/claude-code/overview) running in a sandbox.

    The agent can either use a version of Claude Code installed in the sandbox, or can download a version and install it in the sandbox (see docs on `version` option below for details).

    Use `disallowed_tools` to control access to tools. See [Tools available to Claude](https://docs.anthropic.com/en/docs/claude-code/settings#tools-available-to-claude) for the list of built-in tools which can be disallowed.

    Use the `attempts` option to enable additional submissions if the initial
    submission(s) are incorrect (by default, no additional attempts are permitted).

    Args:
        name: Agent name (used in multi-agent systems with `as_tool()` and `handoff()`)
        description: Agent description (used in multi-agent systems with `as_tool()` and `handoff()`)
        system_prompt: Additional system prompt to append to default system prompt.
        skills: Additional [skills](https://inspect.aisi.org.uk/tools-standard.html#sec-skill) to make available to the agent.
        mcp_servers: MCP servers to make available to the agent.
        bridged_tools: Host-side Inspect tools to expose to the agent via MCP.
            Each BridgedToolsSpec creates an MCP server that makes the specified
            tools available to the agent running in the sandbox.
        disallowed_tools: List of tool names to disallow entirely.
        centaur: Run in 'centaur' mode, which makes Claude Code available to an Inspect `human_cli()` agent rather than running it unattended.
        attempts: Configure agent to make multiple attempts.
        model: Model name to use for Opus and Sonnet calls (defaults to main model for task).
        opus_model: The model to use for `opus`, or for `opusplan` when Plan Mode is active. Defaults to `model`.
        sonnet_model: The model to use for `sonnet`, or for `opusplan` when Plan Mode is not active. Defaults to `model`.
        haiku_model: The model to use for haiku, or [background functionality](https://code.claude.com/docs/en/costs#background-token-usage). Defaults to `model`.
        subagent_model: The model to use for [subagents](https://code.claude.com/docs/en/sub-agents). Defaults to `model`.
        filter: Filter for intercepting bridged model requests.
        retry_refusals: Should refusals be retried? (pass number of times to retry)
        retry_timeouts: Should timeouts be retried? (pass number of times to retry)
        cwd: Working directory to run claude code within.
        env: Environment variables to set for claude code.
        user: User to execute claude code with.
        sandbox: Optional sandbox environment name.
        version: Version of claude code to use. One of:
            - "auto": Use any available version of claude code in the sandbox, otherwise download the current stable version.
            - "sandbox": Use the version of claude code in the sandbox (raises `RuntimeError` if claude is not available in the sandbox)
            - "stable": Download and use the current stable version of claude code.
            - "latest": Download and use the very latest version of claude code.
            - "x.x.x": Download and use a specific version of claude code.
    """
    # resolve centaur
    if centaur is True:
        centaur = CentaurOptions()

    # resolve models
    model = f"inspect/{model}" if model is not None else "inspect"
    opus_model = inspect_model(opus_model)
    sonnet_model = inspect_model(sonnet_model)
    haiku_model = inspect_model(haiku_model)
    subagent_model = inspect_model(subagent_model)

    # resolve skills
    resolved_skills = read_skills(skills) if skills is not None else None

    # resolve attempts
    attempts = AgentAttempts(attempts) if isinstance(attempts, int) else attempts

    async def execute(state: AgentState) -> AgentState:
        # determine port (use new port for each execution of agent on sample)
        MODEL_PORT = "claude_code_model_port"
        port = store().get(MODEL_PORT, 3000) + 1
        store().set(MODEL_PORT, port)

        async with sandbox_agent_bridge(
            state,
            model=model,
            filter=filter,
            retry_refusals=retry_refusals,
            port=port,
            bridged_tools=bridged_tools,
        ) as bridge:
            # ensure claude is installed and get binary location
            claude_binary = await ensure_agent_binary_installed(
                claude_code_binary_source(), version, user, sandbox_env(sandbox)
            )

            # allocate session_id
            session_id = str(uuid.uuid4())

            # base options
            cmd = [
                "--dangerously-skip-permissions",
                "--model",
                model,
            ]

            # add interactive options if not running as centaur
            if centaur is False:
                cmd.extend(
                    [
                        "--print",
                        "--debug",
                        "--verbose",
                    ]
                )

            # system prompt
            system_messages = [
                m.text for m in state.messages if isinstance(m, ChatMessageSystem)
            ]
            if system_prompt is not None:
                system_messages.append(system_prompt)
            if system_messages:
                cmd.extend(["--append-system-prompt", "\n\n".join(system_messages)])

            # mcp servers (combine static configs with bridged tools)
            cmd_allowed_tools: list[str] = []
            all_mcp_servers = list(mcp_servers or []) + bridge.mcp_server_configs
            if all_mcp_servers:
                mcp_server_args, mcp_allowed_tools = resolve_mcp_servers(
                    all_mcp_servers
                )
                cmd.extend(mcp_server_args)
                cmd_allowed_tools.extend(mcp_allowed_tools)

            # add allowed and disallowed tools
            if len(cmd_allowed_tools) > 0:
                cmd.append("--allowed-tools")
                cmd.append(",".join(cmd_allowed_tools))
            if disallowed_tools is not None and len(disallowed_tools) > 0:
                cmd.append("--disallowed-tools")
                cmd.append(",".join(disallowed_tools))

            prompt, has_assistant_response = build_user_prompt(state.messages)

            # resolve sandbox
            sbox = sandbox_env(sandbox)

            # install skills
            if resolved_skills is not None:
                CLAUDE_SKILLS = ".claude/skills"
                skills_dir = (
                    join_path(cwd, CLAUDE_SKILLS) if cwd is not None else CLAUDE_SKILLS
                )
                await install_skills(resolved_skills, sbox, user, skills_dir)

            # define agent env
            agent_env = {
                "ANTHROPIC_BASE_URL": f"http://localhost:{bridge.port}",
                "ANTHROPIC_AUTH_TOKEN": "sk-ant-api03-DOq5tyLPrk9M4hPE",
                "ANTHROPIC_MODEL": model,
                "ANTHROPIC_DEFAULT_OPUS_MODEL": opus_model or model,
                "ANTHROPIC_DEFAULT_SONNET_MODEL": sonnet_model or model,
                "ANTHROPIC_DEFAULT_HAIKU_MODEL": haiku_model or model,
                "CLAUDE_CODE_SUBAGENT_MODEL": subagent_model or model,
                "ANTHROPIC_SMALL_FAST_MODEL": haiku_model or model,
                "CLAUDE_CODE_DISABLE_NONESSENTIAL_TRAFFIC": "1",
                "CLAUDE_CODE_DISABLE_EXPERIMENTAL_BETAS": "1",
                "IS_SANDBOX": "1",
            } | (env or {})

            # centaur mode uses human_cli with custom instructions and bash rc
            if centaur:
                await run_claude_code_centaur(
                    options=centaur,
                    claude_cmd=[claude_binary] + cmd,
                    agent_env=agent_env,
                    state=state,
                )
            else:
                # execute the agent (track debug output)
                debug_output: list[str] = []
                agent_prompt = prompt
                attempt_count = 0
                timeout_count = 0
                while True:
                    # resume previous conversation
                    if has_assistant_response or attempt_count > 0 or timeout_count > 0:
                        agent_cmd = (
                            [claude_binary, "--continue"] + cmd + ["--", agent_prompt]
                        )
                    else:
                        agent_cmd = (
                            [claude_binary, "--session-id", session_id]
                            + cmd
                            + ["--", agent_prompt]
                        )

                    # run agent
                    result = await sbox.exec(
                        cmd=["bash", "-c", 'exec 0<&- "$@"', "bash"] + agent_cmd,
                        cwd=cwd,
                        env=agent_env,
                        user=user,
                        concurrency=False,
                    )

                    # track debug output
                    debug_output.append(result.stdout)
                    debug_output.append(result.stderr)

                    # raise for error
                    if not result.success:
                        # see if this is a timeout and we are retrying timeouts
                        if (
                            "request timed out"
                            in (result.stdout.lower() + result.stderr.lower())
                            and retry_timeouts is not None
                            and timeout_count < retry_timeouts
                        ):
                            timeout_count += 1
                            delay = min(2**timeout_count, 60)
                            trace(
                                f"Retrying timed out request (retry {timeout_count}, waiting {delay} seconds)."
                            )
                            await anyio.sleep(delay)
                            continue

                        raise RuntimeError(
                            f"Error executing claude code agent: {result.stdout}\n{result.stderr}"
                        )

                    # reset timeout counter
                    timeout_count = 0

                    # exit if we are at max_attempts
                    attempt_count += 1
                    if attempt_count >= attempts.attempts:
                        break

                    # score this attempt
                    answer_scores = await score(state)

                    # break if we score 'correct'
                    if attempts.score_value(answer_scores[0].value) == 1.0:
                        break

                    # otherwise update prompt with incorrect message and continue
                    else:
                        if callable(attempts.incorrect_message):
                            if not is_callable_coroutine(attempts.incorrect_message):
                                raise ValueError(
                                    "The incorrect_message function must be async."
                                )
                            agent_prompt = await attempts.incorrect_message(
                                state, answer_scores
                            )
                        else:
                            agent_prompt = attempts.incorrect_message

                # trace debug info
                debug_output.insert(0, "Claude Code Debug Output:")
                trace("\n".join(debug_output))

        return bridge.state

    # return agent with specified name and descritpion
    return agent_with(execute, name=name, description=description)


def resolve_mcp_servers(
    mcp_servers: Sequence[MCPServerConfig],
) -> tuple[list[str], list[str]]:
    # build servers and allowed tools
    mcp_servers_json: dict[str, dict[str, Any]] = {}
    allowed_tools: list[str] = []
    for mcp_server in mcp_servers:
        mcp_servers_json[mcp_server.name] = mcp_server.model_dump(
            exclude={"name", "tools"}, exclude_none=True
        )
        if mcp_server.tools == "all":
            allowed_tools.append(f"mcp__{mcp_server.name}_*")
        elif isinstance(mcp_server.tools, list):
            allowed_tools.extend(
                [f"mcp__{mcp_server.name}__{tool}" for tool in mcp_server.tools]
            )
        else:
            raise ValueError(
                f"Unexpected value for mcp server tools: {mcp_server.tools}"
            )

    # map to cli args
    mcp_config_cmds: list[str] = []
    if len(mcp_servers_json) > 0:
        mcp_config_cmds.append("--mcp-config")
        mcp_config_cmds.append(
            to_json({"mcpServers": mcp_servers_json}, exclude_none=True).decode()
        )

    return mcp_config_cmds, allowed_tools


def inspect_model(model: str | None) -> str | None:
    """Ensure that model name is prefaced with 'inspect/'."""
    if model is not None:
        if model != "inspect" and not model.startswith("inspect/"):
            return f"inspect/{model}"

    return model


async def run_claude_code_centaur(
    options: CentaurOptions,
    claude_cmd: list[str],
    agent_env: dict[str, str],
    state: AgentState,
) -> None:
    instructions = "Claude Code:\n\n - You may also use Claude Code via the 'claude' command.\n - Use 'claude --resume' if you need to resume a previous claude session."

    # build .bashrc content
    agent_env_vars = [f'export {k}="{v}"' for k, v in agent_env.items()]
    claude_config = """echo '{"hasCompletedOnboarding":true,"bypassPermissionsModeAccepted":true}' > "$HOME"/.claude.json"""
    path_config = [
        'mkdir -p "$HOME/.local/bin"',
        'export PATH="$HOME/.local/bin:$PATH"',
        f'ln -sf {claude_cmd[0]} "$HOME/.local/bin/claude"',
    ]
    alias_cmd = shlex.join(claude_cmd)
    alias_cmd = "alias claude='" + alias_cmd.replace("'", "'\\''") + "'"
    bashrc = "\n".join(
        agent_env_vars + path_config + ["", claude_config, "", alias_cmd]
    )

    # run the human cli
    await run_centaur(options, instructions, bashrc, state)



from ._claude_code.claude_code import claude_code
from ._codex_cli.codex_cli import codex_cli
from ._tools.download import AgentBinary, cached_agent_binaries, download_agent_binary
from ._util.centaur import CentaurOptions
from ._util.sandbox import SandboxPlatform

try:
    from ._version import __version__
except ImportError:
    __version__ = "unknown"


__all__ = [
    "claude_code",
    "codex_cli",
    "download_agent_binary",
    "cached_agent_binaries",
    "AgentBinary",
    "SandboxPlatform",
    "CentaurOptions",
    "__version__",
]



from inspect_ai.model import ChatMessage, ChatMessageAssistant, ChatMessageUser


def build_user_prompt(messages: list[ChatMessage]) -> tuple[str, bool]:
    if messages and isinstance(messages[-1], ChatMessageAssistant):
        raise ValueError("Messages input ends with an assistant messages.")

    last_assistant_idx = next(
        (
            i
            for i, m in reversed(list(enumerate(messages)))
            if isinstance(m, ChatMessageAssistant)
        ),
        None,
    )

    has_assistant_response = last_assistant_idx is not None
    start_idx = (last_assistant_idx + 1) if last_assistant_idx is not None else 0

    prompt = "\n\n".join(
        m.text for m in messages[start_idx:] if isinstance(m, ChatMessageUser)
    )

    return prompt, has_assistant_response



import re
from pathlib import Path

from pydantic import BaseModel
from typing_extensions import Literal

from .._util.agentbinary import AgentBinarySource, AgentBinaryVersion
from .._util.appdirs import package_cache_dir
from .._util.download import download_text_file
from .._util.sandbox import SandboxPlatform


def claude_code_binary_source() -> AgentBinarySource:
    cached_binary_dir = package_cache_dir("claude-code-downloads")

    async def resolve_version(
        version: Literal["stable", "latest"] | str, platform: SandboxPlatform
    ) -> AgentBinaryVersion:
        gcs_bucket = await _claude_code_gcs_bucket()
        version = await _claude_code_version(gcs_bucket, version)
        manifest = await _claude_code_manifest(gcs_bucket, version)
        expected_checksum = _checksum_for_platform(manifest, platform)
        download_url = f"{gcs_bucket}/{version}/{platform}/claude"
        return AgentBinaryVersion(version, expected_checksum, download_url)

    def cached_binary_path(version: str, platform: SandboxPlatform) -> Path:
        return cached_binary_dir / f"claude-{version}-{platform}"

    def list_cached_binaries() -> list[Path]:
        return list(cached_binary_dir.glob("claude-*"))

    return AgentBinarySource(
        agent="claude code",
        binary="claude",
        resolve_version=resolve_version,
        cached_binary_path=cached_binary_path,
        list_cached_binaries=list_cached_binaries,
        post_download=None,
        post_install=None,
    )


async def _claude_code_gcs_bucket() -> str:
    INSTALL_SCRIPT_URL = "https://claude.ai/install.sh"
    script_content = await download_text_file(INSTALL_SCRIPT_URL)
    pattern = r'GCS_BUCKET="(https://storage\.googleapis\.com/[^"]+)"'
    match = re.search(pattern, script_content)
    if match is not None:
        gcs_bucket = match.group(1)
        return gcs_bucket
    else:
        raise RuntimeError("Unable to determine GCS bucket for claude code.")


async def _claude_code_version(gcs_bucket: str, target: str) -> str:
    # validate target
    target_pattern = r"^(stable|latest|[0-9]+\.[0-9]+\.[0-9]+(-[^[:space:]]+)?)$"
    if re.match(target_pattern, target) is None:
        raise RuntimeError(
            "Invalid version target (must be 'stable', 'latest', or a semver version number)"
        )

    # resolve target alias if required
    if target in ["stable", "latest"]:
        version_url = f"{gcs_bucket}/{target}"
        version = await download_text_file(version_url)
        return version
    else:
        return target


class PlatformInfo(BaseModel):
    checksum: str
    size: int


class Manifest(BaseModel):
    version: str
    platforms: dict[str, PlatformInfo]


async def _claude_code_manifest(gcs_bucket: str, version: str) -> Manifest:
    manifest_url = f"{gcs_bucket}/{version}/manifest.json"
    manifest_json = await download_text_file(manifest_url)
    return Manifest.model_validate_json(manifest_json)


def _checksum_for_platform(manifest: Manifest, platform: SandboxPlatform) -> str:
    if platform not in manifest.platforms:
        raise RuntimeError(f"Platform '{platform}' not found in manifest.")
    return manifest.platforms[platform].checksum



from inspect_ai.agent import AgentState, human_cli, run
from pydantic import BaseModel, Field


class CentaurOptions(BaseModel):
    """Options for centaur mode."""

    answer: bool | str = Field(default=True)
    """
    Is an explicit answer required for this task or is it scored
    based on files in the container? Pass a `str` with a regex to validate
    that the answer matches the expected format.
    """

    intermediate_scoring: bool = Field(default=False)
    """Allow the human agent to check their score while working."""

    record_session: bool = Field(default=True)
    """Record all user commands and outputs in the sandbox bash session."""


async def run_centaur(
    options: CentaurOptions, instructions: str, bashrc: str, state: AgentState
) -> None:
    agent = human_cli(
        answer=options.answer,
        intermediate_scoring=options.intermediate_scoring,
        record_session=options.record_session,
        instructions=instructions,
        bashrc=bashrc,
    )
    await run(agent, state)



import re
from pathlib import Path
from typing import Literal, NamedTuple

from .._claude_code.agentbinary import claude_code_binary_source
from .._codex_cli.agentbinary import codex_cli_binary_source
from .._util._async import run_coroutine
from .._util.agentbinary import (
    AgentBinarySource,
    download_agent_binary_async,
)
from .._util.sandbox import SandboxPlatform


class AgentBinary(NamedTuple):
    """Agent binary."""

    agent: Literal["claude_code", "codex_cli"]
    """Agent type."""

    version: str
    """Agent version."""

    path: Path
    """"Agent path."""

    def __str__(self) -> str:
        return f"{self.agent} {self.version}: {self.path}"

    def __repr__(self) -> str:
        return f"{self.agent} {self.version}: {self.path}"


class AgentBinaries(list[AgentBinary]):
    def __str__(self) -> str:
        if not self:
            return ""

        max_agent = max(len(b.agent) for b in self)
        max_version = max(len(b.version) for b in self)

        lines = []
        for b in self:
            lines.append(
                f"{b.agent:<{max_agent}}  {b.version:<{max_version}}  {b.path}"
            )
        return "\n".join(lines)

    def __repr__(self) -> str:
        return self.__str__()


def download_agent_binary(
    binary: Literal["claude_code", "codex_cli"],
    version: Literal["stable", "latest"] | str,
    platform: SandboxPlatform,
) -> None:
    """Download agent binary.

    Download an agent binary. This version will be added to the cache of downloaded versions (which retains the 5 most recently downloaded versions).

    Use this if you need to ensure that a specific version of an agent binary is downloaded in advance (e.g. if you are going to run your evaluations offline). After downloading, explicit requests for the downloaded version (e.g. `claude_code(version="1.0.98")`) will not require network access.

    Args:
        binary: Type of binary to download
        version: Version to download ("stable", "latest", or an explicit version number).
        platform: Target platform ("linux-x64", "linux-arm64", "linux-x64-musl", or "linux-arm64-musl")
    """
    match binary:
        case "claude_code":
            source = claude_code_binary_source()
        case "codex_cli":
            source = codex_cli_binary_source()
        case _:
            raise ValueError(f"Unsuported agent binary type: {binary}")

    run_coroutine(download_agent_binary_async(source, version, platform))


def cached_agent_binaries(
    binary: Literal["claude_code", "codex_cli"] | None = None, quiet: bool = False
) -> AgentBinaries:
    """List the agent binaries which have been cached on this system.

    Args:
       binary: Type of binary to list (lists all of if not specified).
       quiet: Do not print the binaries as a side effect

    Returns:
       List of AgentBinary tuples ordered by agent and version (descending).

    """
    if binary is None:
        return AgentBinaries(
            cached_agent_binaries("claude_code") + cached_agent_binaries("codex_cli")
        )

    source = _agent_binary_source(binary)
    binaries = source.list_cached_binaries()

    def parse_name(name: str) -> tuple[str, str, int, int, int]:
        match = re.match(r"([a-z_]+)-(\d+)\.(\d+)\.(\d+)", name)
        if match:
            return (
                match.group(1),  # agent type
                f"{match.group(2)}.{match.group(3)}.{match.group(4)}",  # version string
                int(match.group(2)),  # major
                int(match.group(3)),  # minor
                int(match.group(4)),  # patch
            )
        return ("", "", 0, 0, 0)

    # Sort by type ascending, then version descending
    result = []
    for path in binaries:
        _, version, major, minor, patch = parse_name(path.name)
        if version:
            result.append(
                AgentBinary(
                    agent=binary,
                    version=version,
                    path=path,
                )
            )

    return AgentBinaries(
        sorted(
            result,
            key=lambda x: (
                x.agent,
                -int(x.version.split(".")[0]),
                -int(x.version.split(".")[1]),
                -int(x.version.split(".")[2]),
            ),
        )
    )


def _agent_binary_source(
    binary: Literal["claude_code", "codex_cli"],
) -> AgentBinarySource:
    match binary:
        case "claude_code":
            return claude_code_binary_source()
        case "codex_cli":
            return codex_cli_binary_source()
        case _:
            raise ValueError(f"Unsuported agent binary type: {binary}")



import asyncio
import inspect
from typing import Any, Coroutine, Literal, TypeVar, cast

import nest_asyncio  # type: ignore
import sniffio

from .platform import running_in_notebook

T = TypeVar("T")


def is_callable_coroutine(func_or_cls: Any) -> bool:
    if inspect.iscoroutinefunction(func_or_cls):
        return True
    elif callable(func_or_cls):
        return inspect.iscoroutinefunction(func_or_cls.__call__)
    return False


def run_coroutine(coroutine: Coroutine[None, None, T]) -> T:
    if current_async_backend() == "trio":
        raise RuntimeError("run_coroutine cannot be used with trio")

    if running_in_notebook():
        init_nest_asyncio()
        return asyncio.run(coroutine)
    else:
        try:
            # this will throw if there is no running loop
            asyncio.get_running_loop()

            # initialiase nest_asyncio then we are clear to run
            init_nest_asyncio()
            return asyncio.run(coroutine)

        except RuntimeError:
            # No running event loop so we are clear to run
            return asyncio.run(coroutine)


_initialised_nest_asyncio: bool = False


def init_nest_asyncio() -> None:
    global _initialised_nest_asyncio
    if not _initialised_nest_asyncio:
        nest_asyncio.apply()
        _initialised_nest_asyncio = True


def current_async_backend() -> Literal["asyncio", "trio"] | None:
    try:
        return _validate_backend(sniffio.current_async_library().lower())
    except sniffio.AsyncLibraryNotFoundError:
        return None


def _validate_backend(backend: str) -> Literal["asyncio", "trio"]:
    if backend in ["asyncio", "trio"]:
        return cast(Literal["asyncio", "trio"], backend)
    else:
        raise RuntimeError(f"Unknown async backend: {backend}")



import gzip
import tarfile
from io import BytesIO
from typing import Any, cast


def extract_tarball(tarball_bytes: bytes) -> bytes:
    """Extract the binary from a tar.gz archive."""
    # Open the gzip-compressed tarball
    with BytesIO(tarball_bytes) as tarball_io:
        with gzip.open(tarball_io, "rb") as gz:
            with tarfile.open(fileobj=cast(Any, gz), mode="r") as tar:
                # List all members (should be just one file)
                members = tar.getmembers()
                if len(members) != 1:
                    raise ValueError(
                        f"Expected 1 file in tarball, found {len(members)}"
                    )

                # Extract the binary file
                member = members[0]
                extracted = tar.extractfile(member)
                if extracted is None:
                    raise ValueError(f"Could not extract {member.name}")

                result: bytes = extracted.read()
                return result



from pathlib import Path

from platformdirs import user_cache_path, user_data_path

from .constants import PKG_NAME


def package_data_dir(subdir: str | None) -> Path:
    data_dir = user_data_path(PKG_NAME)
    if subdir:
        data_dir = data_dir / subdir
    data_dir.mkdir(parents=True, exist_ok=True)
    return data_dir


def package_cache_dir(subdir: str | None) -> Path:
    cache_dir = user_cache_path(PKG_NAME)
    if subdir:
        cache_dir = cache_dir / subdir
    cache_dir.mkdir(parents=True, exist_ok=True)
    return cache_dir
